#---FR_exec based of a code of Frank Pilhofer
#---and execlog.tcl from Brent B. Welch

# usage:
# ::FR_exec::Start Command {Time in ms | default: infinity} {Kill the process after finishing | default: 0}
#
# FR_exec opens a pipe for "Command"
# if there no output from "Command" after "Time" we will break
#
# The returnvalue is a list:
# [0 Output_of_the_background_tool] if the backgroundtool was successfully
# [1 Error_reason] if not

package provide FR_exec 0.1

namespace eval ::FR_exec \
{
 namespace export Start
}


#---we open a channel
proc ::FR_exec::Start {Command {Time "infinity"} {Kill 0}} \
{
 variable ExecId; variable ExecContinue; variable ExecChannel; variable ExecData
 variable ExecTimer; variable ExecPid; variable ExecKill; variable ExecLastLine
 
 if {![info exists ExecId]} \
 {
   set ExecId 0
 } else \
 {
   incr ExecId
 }
 
 if {[catch {open "|$Command |& cat"} Channel]} \
 {
   return [list 1 $Channel]
 } else \
 {
   set Id $ExecId
   set ExecChannel($Id) $Channel
   set ExecPid($Id) [lindex [pid $Channel] 0]
   set ExecKill($Id) $Kill
   set ExecData($Id) ""
   set ExecContinue($Id) 0
   set ExecLastLine($Id) ""
  
   if {$Time eq "infinity"} \
   {
     set ExecTimer($Id) $Time
   } \
   else \
   {
     set ExecTimer($Id) [after $Time "::FR_exec::Break $Id"]
   }
  
   fconfigure $Channel -blocking 0 
   fileevent $Channel readable "::FR_exec::FileEvent $Id"
 }
 
 vwait ::FR_exec::ExecContinue($Id)
 
 unset ExecChannel($Id)
 unset ExecContinue($Id)
 
 return [list 0 $ExecData($Id)]
}


#---if the channel readable we execute this
proc ::FR_exec::FileEvent {Id} \
{ 
 variable ExecChannel; variable ExecData; variable ExecLastLine

 if {[eof $ExecChannel($Id)]} \
 {
   ::FR_exec::Stop $Id
 } \
 else \
 {
   gets $ExecChannel($Id) Line
   
   # we don't save more than one empty line
   if {($Line eq "") && ($ExecLastLine($Id) eq "")} \
   {
     # we skip this line
   } \
   else \
   {
     append ExecData($Id) $Line\n
     set ExecLastLine($Id) $Line
   }
 }
}


#---we break after "Time"
proc ::FR_exec::Break {Id} \
{
 variable ExecData
 
 set ExecData($Id) ""
 
 ::FR_exec::Stop $Id
}


#---the stop command after break or after EOF
proc ::FR_exec::Stop {Id} \
{
 variable ExecContinue; variable ExecChannel; variable ExecTimer; variable ExecPid
 variable ExecKill
 
 if {$ExecTimer($Id) ne "infinity"} \
 {
   after cancel $ExecTimer($Id)
 }
 
 fileevent $ExecChannel($Id) readable ""

 if {[catch {close $ExecChannel($Id)} Error]} \
 {
   error "Channel error: $Error"
 }
 
 if {$ExecKill($Id)} \
 {
   catch {exec kill $ExecPid($Id)}
 }
 
 set ExecContinue($Id) 1
}
